from delphai_utils.logging import logging
from delphai_utils.config import get_config
from elasticsearch import AsyncElasticsearch
logging.getLogger('elasticsearch').setLevel(logging.ERROR)
es_host = get_config('elasticsearch.host')
es_index = get_config('elasticsearch.index')
es_username = get_config('elasticsearch.username')
es_password = get_config('elasticsearch.password')
es_port = get_config('elasticsearch.port')
es_address = es_host
if not es_address.startswith('https://'):
  es_address = f'http://{es_host}:{es_port or 80}'
es = AsyncElasticsearch(hosts=[es_address], http_compress=True, http_auth=(es_username, es_password))


from typing import List, Union
from delphai_utils.users import get_user
from grpc import StatusCode
from grpc.aio import ServicerContext
from delphai_utils.logging import logging
import asyncio
from functools import wraps


def get_roles(provided_roles: Union[List[str], ServicerContext]):
  """
  Utility function to retrieve roles from grpc context.
  :param Union[List[str], grpc.aio.ServicerContext] provided_roles:
    a list of roles or the grpc context passed to class methods
  :return list of roles
  :rtype List[str]
  """

  assert isinstance(provided_roles, object) or type(provided_roles) == list
  if type(provided_roles) is not list:
    try:
      user = get_user(provided_roles)
      return user.get('roles') or []
    except KeyError:
      return []
  else:
    return provided_roles


def is_authorized(provided_roles: Union[List[str], ServicerContext],
                  required_roles: List[str],
                  decission_logic: str = 'all',
                  abort: bool = False):
  """
  Makes a decision if user is authorized based on intersection of roles.
  :param Union[List[str], grpc.aio.ServicerContext] provided_roles:
    a list of roles or the grpc context passed to class methods
  :param List[str] required_roles: roles the request must pass
  :param str decission_logic: one of 'all' or 'some'. Default: 'all'
  :param bool abort: if set to true grpc call will abort with PERMISSION_DENIED. Default: False
  :rtype: bool
  """

  assert len(required_roles) != 0
  assert type(required_roles) == list
  assert any([allowed == decission_logic for allowed in ['all', 'some']])
  resolved_roles = get_roles(provided_roles)

  required_roles = set(required_roles)  # in case same role is passed mulitple times
  role_intersection = required_roles.intersection(resolved_roles)
  all_intersects = len(role_intersection) == len(required_roles)
  some_intersects = len(role_intersection) > 0
  _is_authorized = ((decission_logic == 'all' and all_intersects) or (decission_logic == 'some' and some_intersects))
  is_grpc_context = type(provided_roles) is not list
  if abort and not is_authorized and is_grpc_context:
    details = f'Provided roles {resolved_roles} don\'t intersect with required roles {required_roles}.'
    provided_roles.abort(StatusCode.PERMISSION_DENIED, details)
  return _is_authorized


def authorize(required_roles: List[str], decission_logic: str = 'all'):
  """
  Decorator for grpc endpoints to restrict role-based access.
  :param List[str] required_roles: list of strings which roles are required
  :param str decission_logic: one of 'all' or 'some'. Default: 'all'
  :return: wraped function
  :rtype: func
  """
  def wrap(func):
    if not asyncio.iscoroutinefunction(func):

      @wraps(func)
      def wrapped_func(self, request, context, *args, **kwargs):
        if (is_authorized(context, required_roles, decission_logic)):
          return func(self, request, context, *args, **kwargs)
        else:
          details = f"Given roles don't intersect with {required_roles} using decission logic '{decission_logic}'"
          logging.warn(f'403 Forbidden: {details}')
          context.abort(StatusCode.PERMISSION_DENIED, details)

      return wrapped_func
    else:

      @wraps(func)
      async def wrapped_func(self, request, context, *args, **kwargs):
        if (is_authorized(context, required_roles, decission_logic)):
          return await func(self, request, context, *args, **kwargs)
        else:
          details = f"Given roles don't intersect with {required_roles} using decission logic '{decission_logic}'"
          logging.warn(f'403 Forbidden: {details}')
          await context.abort(StatusCode.PERMISSION_DENIED, details)

      return wrapped_func

  return wrap



import grpc
from jose.exceptions import JWTError
from delphai_utils.keycloak import decode_token, PublicKeyFetchError
from delphai_utils.logging import logging
from grpc.aio import Metadata, ServerInterceptor


class AuthenticationInterceptor(ServerInterceptor):
  error_message: str

  def __init__(self):
    async def abort(ignored_request, context):
      await context.abort(grpc.StatusCode.UNAUTHENTICATED, self.error_message)

    self._abort = grpc.unary_unary_rpc_method_handler(abort)

  async def intercept_service(self, continuation, handler_call_details):
    metadata: Metadata = handler_call_details.invocation_metadata
    authorization_header: str = None
    for metadatum in metadata:
      if metadatum[0] == 'authorization':
        authorization_header = metadatum[1]
    if not authorization_header:
      logging.warning('authorization header not specified')
    else:
      if 'Bearer ' not in authorization_header:
        self.error_message = 'Authorization header has the wrong format.'
        logging.error(self.error_message, authorization_header)
        return self._abort
      _, access_token = authorization_header.split('Bearer ')
      try:
        await decode_token(access_token)
      except JWTError as ex:
        self.error_message = f'Error decoding the token: {ex}'
        logging.error(self.error_message)
        return self._abort
      except PublicKeyFetchError as ex:
        self.error_message = f'Error fetching jwk from keycloak: {ex}'
        logging.error(self.error_message)
        return self._abort
    return await continuation(handler_call_details)


from typing import Dict, List, Tuple, Union
from grpc.aio import ServicerContext
from delphai_utils.keycloak import decode_token
import asyncio
import nest_asyncio
import json
from base64 import b64decode

nest_asyncio.apply()


def get_user(context: ServicerContext) -> Dict:
  """
  Get x-delphai-user information.
  :param grpc.aio.ServicerContext context: context passed to grpc endpoints
  :raises: KeyError
  :return x-delphai-user dictionary.
    More information: https://wiki.delphai.dev/wiki/Authorization
  :rtype: Dict
  """

  metadata = dict(context.invocation_metadata())
  if 'authorization' in metadata:
    authorization_header = metadata['authorization']
    _, access_token = authorization_header.split('Bearer ')
    decoded_access_token = asyncio.get_event_loop().run_until_complete(decode_token(access_token))
    user = {
        'https://delphai.com/mongo_user_id': decoded_access_token.get('mongo_user_id'),
        'https://delphai.com/client_id': decoded_access_token.get('mongo_client_id')
    }
    if 'realm_access' in decoded_access_token and 'roles' in decoded_access_token.get('realm_access'):
      roles = decoded_access_token.get('realm_access').get('roles')
      user['roles'] = roles
    if 'group_membership' in decoded_access_token:
      user['groups'] = decoded_access_token['group_membership']
    if 'limited_dataset_group_name' in decoded_access_token:
      user['limited_dataset_group_name'] = decoded_access_token['limited_dataset_group_name']
    if 'name' in decoded_access_token:
      user['name'] = decoded_access_token['name']
    return user
  else:
    user64 = metadata['x-delphai-user']
    user = b64decode(user64)
    return json.loads(user)


def get_groups(context: ServicerContext) -> List[str]:
  """
  Gets groups of calling identity
  :param grpc.aio.ServicerContext context: context passed to grpc endpoints
  :return raw roles passed from keycloak.
    For example this can be ["/delphai/Development"]
  :rtype List[str]
  """

  assert isinstance(context, object)
  try:
    user = get_user(context)
  except KeyError:
    return []
  return user.get('groups') or []


def get_affiliation(context: ServicerContext) -> Tuple[Union[str, None], Union[str, None]]:
  """
  Gets organization and department of user
  :param grpc.aio.ServicerContext context: context passed to grpc endpoints
  :return organization and department as a tuple or None if not affiliated
  :rtype Union[Tuple[str, None], Tuple[str, None]]
  """

  raw_groups = get_groups(context)
  if len(raw_groups) == 0:
    return None, None
  else:
    group_hirarchy = raw_groups[0].split('/')[1:3]
    if len(group_hirarchy) == 1:
      return group_hirarchy.pop(), None
    elif len(group_hirarchy) == 2:
      return tuple(group_hirarchy)
    else:
      return None, None


def get_user_and_client_ids(context: ServicerContext) -> Tuple[str, str]:
  """
  Get user_id and client_id.
  :param grpc.aio.ServicerContext context: context passed to grpc endpoints
  :return user_id and client_id.
    More information: https://wiki.delphai.dev/wiki/Authorization
  :rtype: Tuple[str, str]
  """
  try:
    user = get_user(context)
    user_id = user.get('https://delphai.com/mongo_user_id')
    client_id = user.get('https://delphai.com/client_id')
  except Exception:
    return '', ''
  return user_id, client_id



import logging
from grpc import StatusCode
import grpc
from grpc.aio import ServerInterceptor, AioRpcError
from prometheus_client import (
  Counter,
  Histogram,
  REGISTRY,
  PROCESS_COLLECTOR,
  PLATFORM_COLLECTOR,
)
from prometheus_client.registry import CollectorRegistry
from timeit import default_timer

UNARY = "UNARY"
SERVER_STREAMING = "SERVER_STREAMING"
CLIENT_STREAMING = "CLIENT_STREAMING"
BIDI_STREAMING = "BIDI_STREAMING"
UNKNOWN = "UNKNOWN"

REGISTRY.unregister(PROCESS_COLLECTOR)
REGISTRY.unregister(PLATFORM_COLLECTOR)
REGISTRY.unregister(REGISTRY._names_to_collectors["python_gc_objects_collected_total"])


class MetricsInterceptor(ServerInterceptor):
  registry: CollectorRegistry
  delphai_request_count: Counter
  delphai_response_count: Counter
  delphai_latency_seconds: Histogram

  def __init__(self, registry: CollectorRegistry = REGISTRY) -> None:
    self.registry = registry
    self.delphai_request_count = Counter(
      "delphai_request_count",
      "Total number of RPCs started on the server.",
      ["grpc_service", "grpc_method"],
      registry=registry,
    )
    self.delphai_response_count = Counter(
      "delphai_response_count",
      "Total number of RPCs completed on the server, regardless of success or failure.",
      ["grpc_type", "grpc_service", "grpc_method", "grpc_code"],
      registry=registry,
    )
    self.delphai_latency_seconds = Histogram(
      "delphai_latency_seconds",
      "Histogram of response latency (seconds)",
      ["grpc_type", "grpc_service", "grpc_method", "grpc_code"],
      registry=registry,
    )

  def get_method_type(self, request_streaming, response_streaming):
    """
        Infers the method type from if the request or the response is streaming.
        # The Method type is coming from:
        # https://grpc.io/grpc-java/javadoc/io/grpc/MethodDescriptor.MethodType.html
        """
    if request_streaming and response_streaming:
      return BIDI_STREAMING
    elif request_streaming and not response_streaming:
      return CLIENT_STREAMING
    elif not request_streaming and response_streaming:
      return SERVER_STREAMING
    return UNARY

  def split_method_call(self, handler_call_details):
    """
        Infers the grpc service and method name from the handler_call_details.
        """

    # e.g. /package.ServiceName/MethodName
    parts = handler_call_details.method.split("/")
    if len(parts) < 3:
      return "", "", False

    grpc_service_name, grpc_method_name = parts[1:3]
    return grpc_service_name, grpc_method_name, True

  async def _wrap_rpc_behavior(self, handler, fn):
    """Returns a new rpc handler that wraps the given function"""
    if handler is None:
      return None

    if handler.request_streaming and handler.response_streaming:
      behavior_fn = handler.stream_stream
      handler_factory = grpc.stream_stream_rpc_method_handler
    elif handler.request_streaming and not handler.response_streaming:
      behavior_fn = handler.stream_unary
      handler_factory = grpc.stream_unary_rpc_method_handler
    elif not handler.request_streaming and handler.response_streaming:
      behavior_fn = handler.unary_stream
      handler_factory = grpc.unary_stream_rpc_method_handler
    else:
      behavior_fn = handler.unary_unary
      handler_factory = grpc.unary_unary_rpc_method_handler

    return handler_factory(
      await fn(behavior_fn, handler.request_streaming, handler.response_streaming),
      request_deserializer=handler.request_deserializer,
      response_serializer=handler.response_serializer,
    )

  async def intercept_service(self, continuation, handler_call_details):
    grpc_service_name, grpc_method_name, _ = self.split_method_call(handler_call_details)

    async def metrics_wrapper(behavior, request_streaming, response_streaming):
      async def new_behavior(request_or_iterator, servicer_context):
        response_or_iterator = None
        grpc_code = StatusCode.UNKNOWN
        try:
          start = default_timer()
          grpc_type = self.get_method_type(request_streaming, response_streaming)
          try:
            self.delphai_request_count.labels(
              grpc_service=grpc_service_name,
              grpc_method=grpc_method_name,
            ).inc()
            response_or_iterator = await behavior(request_or_iterator, servicer_context)
            grpc_code = StatusCode.OK
            self.delphai_response_count.labels(
              grpc_type=grpc_type,
              grpc_service=grpc_service_name,
              grpc_method=grpc_method_name,
              grpc_code=grpc_code,
            ).inc()
            return response_or_iterator
          except AioRpcError as ex:
            grpc_code = ex.code()
            logging.error(f"[{ex.code()}] {ex.details()}")
            self.delphai_response_count.labels(
              grpc_type=grpc_type,
              grpc_service=grpc_service_name,
              grpc_method=grpc_method_name,
              grpc_code=grpc_code,
            ).inc()
            raise ex
          finally:
            end = default_timer()
            elapsed = end - start
            logging.info(f"[{grpc_code}] {grpc_service_name}/{grpc_method_name} [{round(elapsed * 1000, 2)}ms]")
            self.delphai_latency_seconds.labels(
              grpc_type=grpc_type,
              grpc_service=grpc_service_name,
              grpc_method=grpc_method_name,
              grpc_code=grpc_code,
            ).observe(max(elapsed, 0))
        except Exception as ex:
          return await continuation(handler_call_details)

      return new_behavior

    if grpc_service_name.startswith("grpc"):
      return await continuation(handler_call_details)
    else:
      handler = await continuation(handler_call_details)
      optional_any = await self._wrap_rpc_behavior(handler, metrics_wrapper)
      return optional_any



import asyncio
from logging import Logger
from aioprometheus.collectors import Histogram
from deepmerge import always_merger
from delphai_utils.logging import default_config, logging
from abc import abstractmethod, abstractproperty
import faust
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, TypeVar, Type
from delphai_utils.config import get_config
from faust.types import CodecArg
from faust.types.models import ModelT
from faust.streams import Stream
from confluent_kafka.admin import AdminClient, NewTopic
from betterproto import Message
from faust.serializers import Codec
from dacite import from_dict
from delphai_utils.config import get_config
import coloredlogs
from aioprometheus import Service, Counter, Summary
from timeit import default_timer

logging.getLogger('aiohttp').setLevel(logging.ERROR)


@dataclass
class Step:
  name: str
  partitions: int
  output: Optional[str]
  tables: Optional[List[str]] = field(default_factory=lambda: [])


T = TypeVar("T")


class FaustProtobufSerializer(Codec):
  def __init__(self, type_class: T, **kwargs):
    self.type_class = type_class
    super(FaustProtobufSerializer, self).__init__(type_class=type_class, **kwargs)

  def _loads(self, s: bytes) -> T:
    return self.type_class().from_json(s.decode('utf-8'))

  def _dumps(self, s: Message) -> bytes:
    return s.to_json().encode('utf-8')


@dataclass
class Metrics:
  inbound_counter: Optional[Counter] = None
  processed_counter: Optional[Counter] = None
  lag_timer: Optional[Summary] = None


class FaustAgent():
  step: str
  step_config: Step
  agent_id: str
  input_topic: faust.TopicT
  output_topic: faust.TopicT
  input_value_type: Type[ModelT]
  output_value_type: Type[ModelT]
  input_value_serializer: CodecArg
  output_value_serializer: CodecArg
  agent: faust.Agent
  logger: Logger
  metrics: Metrics
  metrics_server: Service

  def setup_metrics(self):
    self.metrics = Metrics()
    name = self.agent_id.replace(".", "_").replace('-', '_')
    self.metrics.inbound_counter = Counter(
      f'delphai_{name}_inbound_counter',
      'incoming requests',
      const_labels={
        'step': self.step,
        'pipeline': self.agent.app.conf.id,
      },
    )
    self.metrics.processed_counter = Counter(
      f'delphai_{name}_processed_counter',
      'processed requests',
      const_labels={
        'step': self.step,
        'pipeline': self.agent.app.conf.id,
      },
    )
    self.metrics.lag_timer = Histogram(
      f'delphai_{name}_lag_timer',
      'time to process requests',
      const_labels={
        'step': self.step,
        'pipeline': self.agent.app.conf.id,
      },
    )
    self.metrics_server.register(self.metrics.inbound_counter)
    self.metrics_server.register(self.metrics.processed_counter)
    self.metrics_server.register(self.metrics.lag_timer)

  @abstractproperty
  def step(self) -> str:
    pass

  @abstractproperty
  def input_value_type(self) -> Type[ModelT]:
    pass

  @abstractproperty
  def output_value_type(self) -> Type[ModelT]:
    pass

  @abstractproperty
  def input_value_serializer(self) -> CodecArg:
    pass

  @abstractproperty
  def output_value_serializer(self) -> CodecArg:
    pass

  @abstractmethod
  async def process(self, requests: Stream[Type[ModelT]]):
    pass

  async def on_start(self):
    self.logger.info(f'started agent {self.agent_id}')

  async def after_attach(self):
    pass

  def start_measure(self, labels: Dict[str, str] = {}):
    start = default_timer()
    self.metrics.inbound_counter.inc(labels)
    return start

  def end_measure(self, start: float, labels: Dict[str, str] = {}):
    end = default_timer()
    self.metrics.lag_timer.add(labels, end - start)
    self.metrics.processed_counter.inc(labels)


class FaustApp():
  app: faust.App
  worker: faust.Worker
  logging_config: Dict
  metrics: Service = Service()

  async def on_startup_finished(self):
    self.logger.info(f'started app {self.app._conf.id}')

  async def on_start(self):
    self.logger.info(f'starting app {self.app._conf.id}')

  async def on_shutdown(self):
    self.logger.info(f'shutting down {self.app._conf.id}')

  def on_setup_root_logger(self, logger: logging.Logger, level: int):
    standard_format = self.logging_config['formatters']['standard']
    effective_level = self.logger.getEffectiveLevel()
    logger.setLevel(effective_level)
    coloredlogs.install(
      level=effective_level,
      fmt=standard_format['format'],
      datefmt=standard_format['datefmt'],
      logger=logger,
    )
    self.worker._disable_spinner_if_level_below_WARN(effective_level)
    self.worker._setup_spinner_handler(logger, effective_level)

  def __init__(self, id: str, broker: str) -> None:
    self.app = faust.App(id=id, broker=broker, web_enabled=False)
    loop = asyncio.get_event_loop()
    provided_config = {}
    try:
      provided_config = get_config('logging')
    except Exception:
      pass
    default_config['disable_existing_loggers'] = False
    default_config['formatters']['standard']['format'] = '[%(asctime)s] [%(levelname)s] [%(name)s] %(message)s'
    default_config['loggers']['faust'] = {'level': logging.ERROR}
    default_config['loggers']['aiokafka'] = {'level': logging.ERROR}
    default_config['loggers']['mode'] = {'level': logging.ERROR}
    self.logging_config = always_merger.merge(default_config, provided_config)
    self.worker = faust.Worker(self.app, loop=loop, loglevel=logging.INFO, logging_config=self.logging_config)
    self.worker.on_startup_finished = self.on_startup_finished
    self.worker.on_start = self.on_start
    self.worker.on_shutdown = self.on_shutdown
    self.worker.on_setup_root_logger = self.on_setup_root_logger
    self.logger = logging.getLogger(self.app._conf.id)

  def start(self, agents: List[FaustAgent]):
    try:
      broker = str(self.app._conf.broker[0])
      steps = list(map(lambda s: s['name'], get_config('steps')))
      for step in steps:
        self.create_topic(broker, step)
      tasks = []
      for agent in agents:
        tasks.append(self.attach(agent))
      asyncio.get_event_loop().create_task(self.metrics.start(port=9191))
      self.logger.info('started metrics server on port 9191')
      self.worker.loop.run_until_complete(asyncio.gather(*tasks))
      self.worker.execute_from_commandline()
    except KeyboardInterrupt:
      self.logger.info('keyboard interrupt received')

  def get_step_config(self, step: str) -> Step:
    step_config = next((s for s in get_config('steps') if s['name'] == step), None)
    return from_dict(data_class=Step, data=step_config)

  def create_topic(self, broker: str, step: str):
    step_config = self.get_step_config(step)
    client = AdminClient({'bootstrap.servers': broker.replace('kafka://', '')})
    topics = list(client.list_topics().topics.keys())
    topic_names = [f'{self.app._conf.id}.{step_config.name}']
    for table in step_config.tables:
      table_topic_name = f'{self.app._conf.id}-{table}-changelog'
      topic_names.append(table_topic_name)
    for topic_name in topic_names:
      if topic_name not in topics:
        self.logger.info(f'creating topic {topic_name} with {step_config.partitions} partitions')
        resp = client.create_topics([NewTopic(topic_name, step_config.partitions, 1)])
        resp[topic_name].result()
      else:
        self.logger.info(f'topic {topic_name} already exists')

  async def attach(self, agent: FaustAgent):
    agent.step_config = self.get_step_config(agent.step)
    agent.agent_id = f'{self.app._conf.id}.{agent.step}'
    agent.logger = logging.getLogger(f'{self.app._conf.id}] [{agent.step}')
    agent.input_topic = self.app.topic(
      agent.agent_id,
      value_type=agent.input_value_type,
      value_serializer=agent.input_value_serializer,
    )
    output_topic_name = f'{self.app._conf.id}.{agent.step_config.output}'
    agent.output_topic = self.app.topic(
      output_topic_name,
      value_serializer=agent.output_value_serializer,
    )
    new_agent = self.app.agent(agent.input_topic)(agent.process)
    agent.agent = new_agent
    agent.metrics_server = self.metrics
    agent.setup_metrics()
    await agent.after_attach()
    await new_agent.start()



import time
from typing import Any, Optional, Sequence, Tuple
from grpc.aio import insecure_channel as async_insecure_channel
from grpc import insecure_channel
from random import randint
import grpc


def get_grpc_client(Stub, address: str, is_async: bool = True, options: Optional[Sequence[Tuple[str, Any]]] = {}, max_retries=4, use_retry_interceptor=True):
  sync_interceptors = (
    SyncUnaryUnaryRetryOnRpcErrorClientInterceptor(
      max_attempts=max_retries,
      sleeping_policy=ExponentialBackoff(init_backoff_ms=100, max_backoff_ms=1600, multiplier=2),
      status_for_retry=(grpc.StatusCode.UNAVAILABLE,),
    ),
    SyncUnaryStreamRetryOnRpcErrorClientInterceptor(
      max_attempts=max_retries,
      sleeping_policy=ExponentialBackoff(init_backoff_ms=100, max_backoff_ms=1600, multiplier=2),
      status_for_retry=(grpc.StatusCode.UNAVAILABLE,),
    ),
  )
  async_interceptors = (
    AsyncUnaryUnaryRetryOnRpcErrorClientInterceptor(
      max_attempts=max_retries,
      sleeping_policy=ExponentialBackoff(init_backoff_ms=100, max_backoff_ms=1600, multiplier=2),
      status_for_retry=(grpc.StatusCode.UNAVAILABLE,),
    ),
  )
  if is_async:
    if use_retry_interceptor:
      channel = async_insecure_channel(address, interceptors=async_interceptors)
    else:
      channel = async_insecure_channel(address, options=options)
  else:
    channel = insecure_channel(address, options=options)
    if use_retry_interceptor:
      channel = grpc.intercept_channel(channel, *sync_interceptors)
  client = Stub(channel)
  return client

# Taken from https://github.com/grpc/grpc/issues/19514#issuecomment-531700657
class ExponentialBackoff:
  def __init__(self, *, init_backoff_ms: int, max_backoff_ms: int, multiplier: int):
    self.init_backoff = randint(0, init_backoff_ms)
    self.max_backoff = max_backoff_ms
    self.multiplier = multiplier

  def sleep(self, try_i: int):
    sleep_range = min(
      self.init_backoff * self.multiplier ** try_i, self.max_backoff
    )
    sleep_ms = randint(0, sleep_range)
    time.sleep(sleep_ms / 1000)

class RetryOnRpcErrorClientInterceptor(
    grpc.UnaryUnaryClientInterceptor, grpc.StreamUnaryClientInterceptor
):
  def __init__(
    self,
    *,
    max_attempts: int,
    sleeping_policy: ExponentialBackoff,
    status_for_retry: Optional[Tuple[grpc.StatusCode]] = None,
  ):
    self.max_attempts = max_attempts
    self.sleeping_policy = sleeping_policy
    self.status_for_retry = status_for_retry

  def _intercept_call(self, continuation, client_call_details, request_or_iterator):

    for try_i in range(self.max_attempts):
      response = continuation(client_call_details, request_or_iterator)

      if isinstance(response, grpc.RpcError):

        is_last_attempt = try_i == (self.max_attempts - 1)
        if is_last_attempt:
          return response

        is_not_retryable_status = (
          self.status_for_retry
          and response.code() not in self.status_for_retry
        )
        if is_not_retryable_status:
          return response

        self.sleeping_policy.sleep(try_i)
      else:
        return response

  def intercept_unary_unary(self, continuation, client_call_details, request):
    return self._intercept_call(continuation, client_call_details, request)

  def intercept_stream_unary(
    self, continuation, client_call_details, request_iterator
  ):
    return self._intercept_call(continuation, client_call_details, request_iterator)

class SyncUnaryUnaryRetryOnRpcErrorClientInterceptor(
  RetryOnRpcErrorClientInterceptor, grpc.UnaryUnaryClientInterceptor
):
  def intercept_unary_unary(self, continuation, client_call_details, request):
      return self._intercept_call(continuation, client_call_details, request)

class SyncUnaryStreamRetryOnRpcErrorClientInterceptor(
  RetryOnRpcErrorClientInterceptor, grpc.UnaryStreamClientInterceptor
):
  def intercept_unary_stream(self, continuation, client_call_details, request):
      return self._intercept_call(continuation, client_call_details, request)

class AsyncUnaryUnaryRetryOnRpcErrorClientInterceptor(
  RetryOnRpcErrorClientInterceptor, grpc.aio.UnaryUnaryClientInterceptor
):
  def intercept_unary_unary(self, continuation, client_call_details, request):
      return self._intercept_call(continuation, client_call_details, request)
